# # basicMLpy.utils module
import numpy as np

def check_for_intercept(x):
    """
    Checks if a given array has an intercept column,i.e. a column of ones. Adds the intercept if the array doesn't have it.
    Inputs:
        x: array
            input array of input points.
    Returns:
        x: array
            outputs the modified array of input points.

    """
    if(len(x.shape) != 1):
        if(not (x[:,0] == 1).all()):
            print("Array of inputs was passed without intercept.\nAdding intercept...")
            ones = np.ones((len(x),1))
            x = np.hstack((ones,x))
            return x 
        else:
            return x 
    else:
        print("Array of inputs was passed without intercept. Adding intercept...")
        ones = np.ones((len(x)))
        x = np.vstack((ones,x)).T
        return x 

def residuals(y,prediction,loss_func):
    """
    Calcultes the generalized(or partial) residuals between the prediction and the output. The residuals are defined to be the negative value of the gradient of the loss function w.r.t. prediction.
    Inputs:
        y: array
            input the array of outputs.
        prediction: array
            input the array of predictions.
        loss_func: string
            input the string that identifies the loss function to use when calculating the residuals; loss_func can be 'mse'(Mean Squared Error), 'mae'(Mean Absolute error).
    Returns:
        gradient: ndarray
            outputs the residuals between the prediction and the output.

    """
    possible_loss_funcs = ['mse','mae']
    assert loss_func in possible_loss_funcs

    if loss_func == possible_loss_funcs[0]:
        return (y - prediction)
    else:
        return (2*(y - prediction) - 1).reshape((-1,1))

def optimal_gamma(y,prediction,loss_func):
    """
    Calculates the optimal value for the gamma constant based on a certain loss_func.
    Inputs:
        y: array
            input the array of outputs/targets.
        prediction: array
            input the array of predictions.
        loss_func: string
            input the string that identifies the loss function to be used; loss_func can be: 'mse'(Mean Squared Error) or 'mae'(Mean Absolute Error)
    Returns:
        np.median(res) or np.mean(res): float or ndarray
            outputs the optimal value for the gamma parameter given the prediction and outputs.
    """
    assert loss_func in ["mse","mae"]

    if loss_func == 'mae':
        res = y - prediction
        return np.median(res)  

    elif loss_func == 'mse':
        res = y - prediction
        return np.mean(res)

def split_indices(x,n_folds):
    """
    Splits the indices of a given array into n folds.
    Inputs:
        x: array
            input the array of input points.
        n_folds: int
            input the number of folds to create.
    Returns:
        fold_idxs: list of lists of int
            outputs a list of folds. each fold is a list of indices.
    """
    try:
        assert len(x) % n_folds == 0
    except:
        raise ValueError("Number of samples must be divisible by the number of folds!")

    fold_idxs = []
    indices = list(range(len(x)))
    np.random.shuffle(indices)
    fold_size = int(len(x)/n_folds)

    for i in range(n_folds):
        if(i == 0):
            fold_idxs.append(indices[:((i+1) * fold_size)])
        elif((i != n_folds - 1)):
            fold_idxs.append(indices[((i) * fold_size):((i+1) * fold_size) ])
        else:
            fold_idxs.append(indices[((i) * fold_size):])

    return fold_idxs
    
def euclidean_distance(x1, x2):
    """
    Calculates the euclidian distance between two points.
    Inputs:
        x1: array_like
            input a point(row vector or just a float/int number) to be used in the calculation.
        x2: array_like
            input a point(row vector or just a float/int number) to be used in the calculation.
    Returns: 
        np.sqrt(distance): float
            outputs the euclidean distance between x1 and x2.
    """
    distance = 0
    for i in range(len(x1)-1):
        distance += (x1[i] - x2[i])**2
    return np.sqrt(distance)


# # basicMLpy.classification module
import numpy as np
from scipy import linalg
from basicMLpy.utils import check_for_intercept

#CLASSIFICATION#

def probability_k1(row,parameters):
    """
    Calculates Pr(G = 0|X = x).
    Inputs:
        row: array_like
             input array(row vector), represents a row of the matrix of input points.
        parameters: array
             input the column vector of predictors.
    Returns:
        result: float
              outputs Pr(G = 0|X = x).
    """
    result = np.exp(row @ parameters)/(1 + np.exp(row @ parameters))
    return result 

def probability_vector(dataset,parameters):
    """
    Calculates the vector of probabilities. Each element represents the probability of a given x belonging to class k = 0.
    Inputs:
        dataset: array
             input array of input points, already expected to include the intercept. 
        parameters: array
             input the column vector of predictors.
    Returns: 
        p: array
             outputs the p vector of probabilities.
    """
    p = np.zeros((len(dataset),1))
    for i in range(len(dataset)):
        p[i] = probability_k1(dataset[i,:],parameters)
    return p 

def weight_matrix(dataset,parameters):
    """
    Calculates the diagonal matrix of weights, defined by: W[i,i] = Pr(G = 0|X = x_i) * (1 - Pr(G = 0|X = x_i)).
    Inputs:
        dataset: array
             input array of input points, already expected to include the intercept.
        parameters: array
             input the column vector of predictors.
    Outputs:
        w: array
             outputs a diagonal matrix NxN(N being the number of train samples).
    """
    w = np.eye(len(dataset))
    for i in range(len(dataset)):
        w[i,i] = probability_k1(dataset[i,:],parameters) * (1 - probability_k1(dataset[i,:],parameters))
    return w 

def newton_step(dataset,y,n_iter):
    """
    Calculates the newton step for a given array of input points and it's corresponding vector of output points.
    Inputs:
        dataset: array
             input array of input points, already expected to include the intercept.
        y: array
             input array of output points.
        n_iter: int
            Input the number of iterations for the IRLS algorithm. The algorithm is pretty expensive, so I recommend starting with small values(by experience 15 seems to be a good guess) and then start slowly increasing it untill convergence.
    Returns:
        theta: array
            outputs array of predictors/parameters calculated by the algorithm.   
    """
    theta = np.zeros((np.size(dataset,1),1))
    for i in range(n_iter):
        z = dataset @ theta + np.linalg.pinv(weight_matrix(dataset,theta)) @ (y - probability_vector(dataset,theta))
        theta = np.linalg.pinv(dataset.T @ weight_matrix(dataset,theta) @ dataset) @ dataset.T @ weight_matrix(dataset,theta) @ z 
    return theta


def one_vs_all_default(x,y,k,n_iter):
    """
    Fits a one-vs-all classification model on a given dataset of k > 2 classes.
    Inputs:
        x: array
            input array of input points to be used as training set
        y: array
            input array of output points.
        n_iter: int
            Input the number of iterations for the IRLS algorithm. The algorithm is pretty expensive, so I recommend starting with small values(by experience 15 seems to be a good guess) and then start slowly increasing it untill convergence.
    Returns:
        theta: array
            outputs array of predictors/parameters calculated by the algorithm.

    """
    if k <= 2:
        raise ValueError("Number of classes must be bigger than two!")
    else:
        x = check_for_intercept(x)
        prob_index = np.zeros((len(x),k))
        theta = np.zeros((np.size(x,1),k))
        target = y.copy()
        for i in range(k):
            for w in range(len(y)):
                if y[w] == i:
                    target[w] = 1
                else:
                    target[w] = 0

            parameters = newton_step(x,target.reshape((-1,1)),n_iter)
            theta[:,i] = parameters[:,0]
        return theta


class IRLSClassifier:
    """
    Iteratively Reweighted Least Squares algorithmn for classification, that can solve both binary and multiclass problems.
    Methods:
        fit(X,y) -> Performs the IRLS algorithm on the training set(x,y).
        predict(x) -> Predict the class for X.
        get_prob -> Predict the probabilities for X.
        parameters() -> Returns the calculated parameters for the linear model.

    """
    def __init__(self,k,n_iter=15):
        """
        Initialize self.
        Inputs:
            k: int
                input the number k of classes associated with the dataset. 
            n_iter: int
                Input the number of iterations for the IRLS algorithm. The algorithm is pretty expensive, so I recommend starting with small values(by experience 15 seems to be a good guess) and then start slowly increasing it untill convergence; default is set to 15.
        """
        self.k = k 
        if self.k <= 1:
            raise ValueError("Insert a valid number of classes!")
        self.n_iter = n_iter

    def fit(self,x,y):
        """
        Fits the classification model on the dataset.
        Inputs:
            x: array
                input array of input points to be used as training set, without the intercept(raw data).
            y: array
                input array of output points, usually a column vector with same number of rows as x.
        Functionality:
            stores all information from a given function into self.result. this statement will be used later in other functions.            
        """
        self.x = check_for_intercept(x)
        self.y = y 
        if self.k == 2:
            self.parameters = newton_step(self.x,self.y,self.n_iter)
        elif self.k >= 3 : 
            self.parameters = one_vs_all_default(self.x,self.y,self.k,self.n_iter)
    def predict(self,x):
        """
        Gives the prediction made by a certain function based on the input.
        Inputs:
            x: float or array
                input the array of input points or a single input point.
        Returns:
            predictions: float or array
                outputs the prediction made by the classification algorithm.
        """
        x = check_for_intercept(x)
        if self.k == 2:
            predictions = np.round(probability_vector(x,self.parameters))
            return predictions
        else:
            probability_matrix = np.zeros((len(x),self.k))
            predictions = np.zeros((len(x),1))
            for j in range(self.k):
                prob = probability_vector(x,self.parameters[:,j])
                probability_matrix[:,j] = prob[:,0]
            for i in range(len(predictions)):
                predictions[i,0] = np.argmax(probability_matrix[i,:])   
            return predictions 
    def get_prob(self,x):
        """
        Gives the predicted probabilities(without rounding up the results) made by the IRLS algorithm.
        Inputs:
            x: float or array
                input the array of input points or a single input point.
        Returns:
            predictions: float or array
                outputs the prediction(probability) made by the classification algorithm.       
        """
        x = check_for_intercept(x)
        if self.k == 2:
            predictions = probability_vector(x,self.parameters)
            return predictions
        else:
            probability_matrix = np.zeros((len(x),self.k))
            predictions = np.zeros((len(x),1))
            for j in range(self.k):
                prob = probability_vector(x,self.self.parameters[:,j])
                probability_matrix[:,j] = prob[:,0]
            for i in range(len(predictions)):
                predictions[i] = probability_matrix[i,np.argmax(probability_matrix[i,:])]   
            return predict
    def parameters(self):
        """
        Gives the parameters calculated by the classification function.
        Returns:
            self.result[0]: array
                outputs the array of parameters calculated by the classification function.
        """
        return self.parameters

##############



# # basicMLpy.ensemble module
import numpy as np 
from sklearn.tree  import DecisionTreeClassifier
from sklearn.tree  import DecisionTreeRegressor
from sklearn.tree._tree import TREE_LEAF
from basicMLpy.utils import residuals, optimal_gamma
from basicMLpy.loss_functions import misclassification
#BAGGING
class RandomForestRegressor:
    """
    Class of the Random Forest Regressor Model.
    Methods:
        fit(X,y) -> Performs the random forests algorithm on the training set(x,y).
        predict(x) -> Predict regression value for X.
    
    """
    def __init__(self,n_estimators,max_features =1/3,max_depth=None,criterion='mse',random_state=None):
        """
        Initialize self.
        Inputs:
            n_estimators: int
                input the number of trees to grow.
            max_depth: int
                input the maximum depth of the tree; default is set to None.
            criterion: string
                input string that identifies the criterion to be used when deciding how to split each node. criterion can be: 'mse', 'friedman_mse' and 'mae' .default is set to 'mse'.
            max_features = string or int/float:
                input string or int/float that identifies the maximum number of features to be used when splitting each decision tree; if string can be: 'sqrt' or 'log2'; if int max_fetures will be the maximum number of features; if float the maximum number of features will be int(max_features * n_features); default is set to 1/3.             
            random_state: int
                input the random_state to be used on the sklearn DecisionTreeClassifier. default is set to None.
        """
        self.n_estimators = n_estimators
        self.criterion = criterion
        self.random_state = random_state
        self.max_features = max_features
        self.max_depth = max_depth
        possible_criterion = ['mse', 'friedman_mse', 'mae']
        assert self.criterion in possible_criterion 

    def fit(self,X,y):
        """
        Fits the RandomForestRegressor model.
        Inputs:
            X: array
                input array of input points.
            y: array
                input array of output points.
        """
        self.input_train = X
        self.output_train = y.reshape((-1,1))
        self.trained_trees_list = list()

        for i in range(self.n_estimators):
            train_inds = np.random.choice(self.input_train.shape[0],self.input_train.shape[0],True)
            model_tree = DecisionTreeRegressor(criterion = self.criterion, random_state = self.random_state,max_features = self.max_features,max_depth=self.max_depth)
            model_tree = model_tree.fit(self.input_train[train_inds,:],self.output_train[train_inds,:])
            self.trained_trees_list.append(model_tree)

    def predict(self,x):
        """
        Predicts the value of a given group of inputs points based on the trained trees.
        Inputs:
            x: array_like
                input the input point/array to be predicted.
        Returns:
            final_prediction: array_like
                outputs the value prediction of the input made by the random forest model.
        """
        indiv_predictions = np.array([self.classifier.predict(x) for self.classifier in self.trained_trees_list]).T
        final_prediction = np.zeros((indiv_predictions.shape[0],))

        for i in range(indiv_predictions.shape[0]):
            final_prediction[i] = np.sum(indiv_predictions[i][:])/indiv_predictions.shape[1]
        
        return final_prediction

class RandomForestClassifier:
    """
    Class of the Random Forest Classifier Model.
    Methods:
        fit(X,y) -> Performs the random forests algorithm on the training set(x,y).
        predict(x) -> Predict class for X.
    
    """
    def __init__(self,n_estimators,n_classes,max_depth=None,criterion='gini',random_state=None,max_features='sqrt'):
        """
        Initialize self.
        Inputs:
            n_estimators: int
                input the number of trees to grow.
            n_classes: int
                input the number of classes of the classification task.
            max_depth: int
                input the maximum depth of the tree; default is set to None.
            criterion: string
                input string that identifies the criterion to be used when deciding how to split each node. criterion can be: 'gini' or 'entropy' .default is set to 'gini'.
            max_features = string or int/float:
                input string or int/float that identifies the maximum number of features to be used when splitting each decision tree; if string can be: 'sqrt' or 'log2'; if int max_fetures will be the maximum number of features; if float the maximum number of features will be int(max_features * n_features); default is set to 'sqrt'.                   
            random_state: int
                input the random_state to be used on the sklearn DecisionTreeClassifier. default is set to None.
        """
        self.n_estimators = n_estimators
        self.criterion = criterion
        self.random_state = random_state
        self.n_classes = n_classes
        self.max_features = max_features
        self.max_depth = max_depth

        possible_criterion = ['gini','entropy']
        assert self.criterion in possible_criterion 

    def fit(self,X,y):
        """
        Fits the RandomForestClassifier model.
        Inputs:
            X: array
                input array of input points.
            y: array
                input array of output points.
        """
        self.input_train = X
        self.output_train = y.reshape((-1,1))
        self.trained_trees_list = list()

        for i in range(self.n_estimators):
            train_inds = np.random.choice(int(self.input_train.shape[0]/2),int(self.input_train.shape[0]/2),False)
            model_tree = DecisionTreeClassifier(criterion = self.criterion, random_state = self.random_state,max_features = self.max_features,max_depth=self.max_depth,)
            model_tree = model_tree.fit(self.input_train[train_inds,:],self.output_train[train_inds,:])
            self.trained_trees_list.append(model_tree)

    def predict(self,x):
        """
        Predicts the class of a given group of inputs points based on the trained trees. 
        Inputs:
            x: array_like
                input the input point/array to be predicted.
        Returns:
            final_prediction: array_like
                outputs the class prediction of the input made by the random forest model.
        """
        indiv_predictions = np.array([self.classifier.predict(x) for self.classifier in self.trained_trees_list]).T
        final_prediction = np.zeros((indiv_predictions.shape[0],))
        counter_vec = np.zeros((self.n_classes,1))

        for i in range(indiv_predictions.shape[0]):
            for j in range(indiv_predictions.shape[1]):
                counter_vec[indiv_predictions[i][j]] += 1
                final_prediction[i] = np.argmax(counter_vec)
            counter_vec = np.zeros((counter_vec.shape))

        return final_prediction

#BOOSTING
class GBRegressor:
    """
    GradientBoost algorithm for supervised learning, that can fit regression problems.
    Methods:
        fit(X,y) -> Performs the gradient boosting algorithm on the training set(x,y).
        predict(x) -> Predict value for X.
    
    """
    def __init__(self,n_estimators,loss_func,max_features=None,max_depth=None,random_state=None):
        """
        Inilialize self.
        Inputs:
            n_estimators: int
                input the number of trees to grow.
            loss_func: string
                input the string that identifies the loss function to use when calculating the residuals; loss_func can be  'mse'(Mean Squared Error), 'mae'(Mean Absolute error).
            max_features = string or int/float:
                input string or int/float that identifies the maximum number of features to be used when splitting each decision tree; if string can be: 'sqrt' or 'log2'; if int max_fetures will be the maximum number of features; if float the maximum number of features will be int(max_features * n_features); default is set to 'sqrt'.                   
            max_depth: int
                input the maximum depth of the tree; default is set to None.
            random_state: int
                input the random_state to be used on the sklearn DecisionTreeClassifier; default is set to None.
            
                
        """
        possible_params = ['mse', 'mae']
        assert n_estimators > 0
        assert loss_func in possible_params
        if max_depth != None:
            assert max_depth >= 1
            
        self.n_estimators = n_estimators
        self.loss_func = loss_func
        self.random_state = random_state
        self.max_depth = max_depth
        self.max_features = max_features

    def fit(self,X,y):
        """
        Fits the GradientBooster model.
        Inputs:
            X: array
                input array of input points.
            y: array
                input array of output points.

        """
        self.input_train = X
        self.output_train = y
        self.trained_trees_list = list()
        self.gammas = list()
        self.gamma_0 = optimal_gamma(self.output_train,np.zeros(self.output_train.shape[0]),self.loss_func)
        raw_pred = np.ones((self.output_train.shape[0])) * self.gamma_0

        for m in range(self.n_estimators):
            curr_residuals = residuals(self.output_train,raw_pred,self.loss_func)
            model = DecisionTreeRegressor(criterion = self.loss_func, random_state = self.random_state,max_depth=self.max_depth,max_features=self.max_features)
            tree = model.fit(self.input_train,curr_residuals)
            terminal_regions = tree.apply(self.input_train)
            gamma = np.zeros((len(tree.tree_.children_left)))

            for leaf in np.where(tree.tree_.children_left == TREE_LEAF)[0]:
                mask = np.where(terminal_regions == leaf)
                gamma[leaf] = optimal_gamma(self.output_train[mask],raw_pred[mask],self.loss_func)

            raw_pred += gamma[terminal_regions]
            self.trained_trees_list.append(tree)
            self.gammas.append(gamma)

                
    def predict(self,x):
        """
        Predicts the value or class of a given group of inputs points based on the trained trees.
        Inputs:
            x: array_like
                input the input point/array to be predicted.
        Returns:
            final_prediction: array_like
                outputs the class/value prediction of the input made by the gradient booster model.
        """
        raw_pred = np.ones(x.shape[0])* self.gamma_0

        for tree,gamma in zip(self.trained_trees_list,self.gammas):
            terminal_regions = tree.apply(x)
            raw_pred += gamma[terminal_regions]

        return raw_pred         


class AdaBoostClassifier:
    """
    AdaBoost algorithm for weak classifiers, that can fit discrete classification problems.
    Methods:
        fit(x,y) -> Performs the boosting algorithm on the training set(x,y).
        predict(x) -> Predict class for X.
        get_tree_weights() -> Returns the weights for each of the n_iter trees generated during the boosting task.
    """
    def __init__(self,n_estimators):
        """
        Initialize self.
        Inputs:
            n_estimators: int
                input the number of trees(stumps) to grow.
        """
        self.n_estimators = n_estimators

    def fit(self,X,y):
        """
        Fits the AdaBooster on a given dataset.
        Inputs:
            X: array
                input the array of input points
            y: array
                input the array of output points, with y E {-1,1}   
        """
        self.input_train = X 
        self.output_train = y 
        alphas = np.zeros((self.n_estimators,1)) 
        predictions_train = np.zeros((len(self.output_train),self.n_estimators)) 
        staged_weights = np.zeros((len(self.output_train),self.n_estimators)) 
        weighted_error = np.zeros((self.n_estimators,1)) 
        stumps = list()

        for i in range(len(self.output_train)):
            staged_weights[i,0] = 1/len(self.output_train)

        for m in range(self.n_estimators):

            curr_weights = staged_weights[:,m]
            stump = DecisionTreeClassifier(max_depth=1, max_leaf_nodes=2)
            stump = stump.fit(self.input_train,self.output_train,sample_weight=curr_weights)

            predictions_train[:,m] = stump.predict(self.input_train)
            curr_weights = curr_weights.reshape((-1,1))
            weighted_misclassification = curr_weights * misclassification(self.output_train,predictions_train[:,m])
            weighted_error[m] = np.sum(weighted_misclassification)/np.sum(curr_weights)
            alphas[m] = np.log((1 - weighted_error[m])/weighted_error[m])
            curr_weights = curr_weights * np.exp(alphas[m] * misclassification(self.output_train,predictions_train[:,m]))
            curr_weights = curr_weights.reshape((-1,))

            if m + 1 < self.n_estimators:
                staged_weights[:,m+1] = curr_weights
            stumps.append(stump)    

        self.tree_weights = alphas
        self.stumps_list = stumps
        self.weighted_error = weighted_error

    def predict(self,x):
        """
        Makes a prediction based on the weights and classifiers calculated by the Adabooster
        Inputs:
            x: array_like
                input the array of input points
        Returns:
            self.predict: array_like
                outputs the array of predictions
        """   
        indiv_predictions = np.array([self.classifier.predict(x) for self.classifier in self.stumps_list])
        prediction = indiv_predictions.T @ self.tree_weights         
        prediction = np.sign(prediction)

        return prediction

    def get_tree_weights(self):
        """
        Gives the weights calculated by the AdaBooster
        """
        return self.tree_weights




"""
basicMLpy 1.0.2
=================


Provides
    1. Simple and effective machine learning algorithms for classification and regression
    2. Easy-to-use functions that allow the user to expand on the package's capabilities
    3. A Cross-Validation algorithm for the functions presented by the package
=================

Modules
    1.basicMLpy.regression
    2.basicMLpy.classification
    3.basicMLpy.nearest_neighbors
    4.basicMLpy.ensemble
    5.basicMLpy.model_selection
    6.basicMLpy.loss_functions
    7.basicMLpy.utils
=================


For more information on the package, please visit my Github page or the package page at pypi
Thanks for using the package and feel free to contact me for suggestions! 

"""


# # basicMLpy.cross_validation module
import numpy as np 
import itertools
from basicMLpy.utils import check_for_intercept, split_indices


class CrossValidation:
    """
    Class that performs the cross validation given a certain function.
    Methods:
        fit(X,y) -> Performs the cross validation algorithm on the training set(x,y).
        cv_scores() -> Gives the cross validation scores for the training set.
        expected_generalization_error() -> Gives the predicted generalization(out of sample) test error.
        get_cv_estimators() -> Returns all the estimators trained during the cross validation. Requires bool return_estimator to be set to True.
    """
    def __init__(self,estimator,loss_function,n_folds,return_estimator=False):
        """
        Initialize self.
        Inputs:
            estimator: estimator object implementing fit
                input the estimator to be used in the cross validation algorithm
            loss_functions: loss function object return loss betweet targets and predictions
                input the loss function to be used when calculating the cross validation algorithm
            n_folds: int
                input number of folds to be created. must be bigger than two.
            return_estimator: bool, default=False
                Whether to return the estimators fitted on each fold.
        """
        self.estimator = estimator
        self.loss_function = loss_function
        self.n_folds = n_folds
        self.return_estimator = return_estimator

    def fit(self,x,y):
        """
        Performs the cross-validation on a given dataset.
        Inputs:
            x: array
                input array of input points, without the intercept(raw data).
            y: array
                input array of output points.
        """
        self.test_scores = []
        self.train_scores = []
        self.folds = split_indices(x,self.n_folds)
        if self.return_estimator:
            self.cv_estimators = []
        for curr_fold in range(len(self.folds)):
            curr_train_set = list(itertools.chain.from_iterable([x for i,x in enumerate(self.folds) if i != curr_fold]))
            self.estimator.fit(x[curr_train_set],y[curr_train_set])
            self.test_scores.append(self.loss_function(self.estimator.predict(x[self.folds[curr_fold]]),y[self.folds[curr_fold]]))
            self.train_scores.append(self.loss_function(self.estimator.predict(x[curr_train_set]),y[curr_train_set]))
            if self.return_estimator:
                self.cv_estimators.append(self.estimator)

    def scores(self):
        """
        Gives the calculated cross-validation scores for the dataset.
        Returns:
            scores_dict: dict
                outputs a dict of arrays. This dict has two keys: 'train_scores' and 'test_scores', each mapping to an array of the respective scores.
        """
        scores_dict = {}
        scores_dict['train_scores'] = self.train_scores
        scores_dict['test_scores'] = self.test_scores
        return scores_dict   

    def expected_generalization_error(self):
        """
        Calculates the expected test error of the model, that by definition is the average of the sum of the cross-validation error found by the algorithm.
        Returns:
            self.error: float
                Outputs the expected test error of the model.
        """
        return sum(self.scores)/self.n_folds

    def get_cv_estimators(self):
        """
        Returns all of the estimators fitted in the cross validation.
        Returns:
            self.cv_estimators: list of estimator objects
                Outputs the estimators fitted in the cross validation.
        """
        try:
            assert self.return_estimator == True
        except:
            raise ValueError("return_estimator must be set to True in order to use this method!")
        return self.cv_estimators

###


# # basicMLpy.regression module
import numpy as np 
from basicMLpy.utils import check_for_intercept
from scipy import linalg


def qr_regression(x,y):
	"""
	Calculates the predictors for a linear regression model, using QR decomposition.
	Inputs:
		x: array
			input array of input points, with the intercept.
		y: array
			input array of output points, usually a column vector with same number of rows as x.
	Returns:
		theta: array
			outputs the array of predictors for the regression model.
	"""
	x = check_for_intercept(x)
	q, r = np.linalg.qr(x) 
	b = q.T @ y
	theta = linalg.solve_triangular(r,b)
	return theta

def ridge_regression(x,y,reg_const):
	"""
	Fits a ridge linear regression model on a given dataset.
	Inputs:
		x: array
			input array of input points to be used as training set, with the intercept.
		y: array
			input array of output points, usually a column vector with same number of rows as x.
		reg_const: float
			input the value for the penalizing constant(lambda) used by the ridge algorithm.
	Returns:
		theta: array
			outputs the array of predictors for the regression model.
	"""
	x = check_for_intercept(x)
	identity = np.eye(np.size(x,1))
	theta = np.linalg.inv(x.T @ x + reg_const * identity) @ x.T @ y 
	return theta

class LinearRegression:
	"""
	Class of the Linear Regression model.
	Methods:
		fit(x,y) -> Performs the linear regression algorithm on the training set(x,y).
		predict(x) -> Predict value for X.
		parameters() -> Returns the calculated parameters for the linear model.
	"""
	def fit(self,x,y):
		"""
		Fits one of the regression models on the dataset.
		Inputs:
			x: array
				input array of input points.
			y: array
				input array of output points.
		Functionality:
			stores all information from a given function into self.result. this statement will be used later in other functions.
		"""
		self.parameters = qr_regression(x,y)
	def predict(self,x):
		"""
		Gives the prediction made by a certain function based on the input.
		Inputs:
			x: float or array
				input the array of input points or a single input point.
		Returns:
			self.predict: float or array
				outputs the prediction made by the classification algorithm
		"""
		x = check_for_intercept(x)
		predict = x @ self.parameters
		return predict
	def get_parameters(self):
		"""
		Gives the parameters calculated by the regression function.
		Returns:
			self.result[0]: array
				outputs the array of parameters calculated by the regression function.
		"""
		return self.parameters

class RidgeRegression(LinearRegression):
	"""
	Class of the ridge regression model. Inherits from the LinearRegression class.
	Methods:
		fit(x,y) -> Performs the linear regression algorithm on the training set(x,y).
		predict(x) -> Predict value for X.
		parameters() -> Returns the calculated parameters for the linear model.
	"""
	def __init__(self,reg_lambda):
		self.reg_lambda = reg_lambda

	def fit(self,x,y):
		"""
		Fits one of the regression models on the dataset.
		Inputs:
			x: array
				input array of input points, without the intercept(raw data).
			y: array
				input array of output points, usually a column vector.
			reg_lambda: float
				input value that specifies the regularization constant to be used in the ridge regresssion;
		"""
		self.parameters = ridge_regression(x,y,self.reg_lambda)




# # basicMLpy.nearest_neighbors module
import numpy as np
from scipy import linalg 
from basicMLpy.utils import euclidean_distance
def get_neighbors(x, row_num, n_neighbors):
    """
    Gets the K nearest neighbors relative to a point in the dataset.
    Inputs:
        x: array
            input the array of input points.
        row_num: int
            input the index(row number) of the desired point to calculate the k nearest neighbors.
        n_neighbors: int
            input the number of neighbors to be calculated.
    Returns:
        neighbors: array
            outputs the array of the k nearest neighbors to the inputed point.
    """
    distances = list()
    for row in x:
        dist = euclidean_distance(x[row_num,:], row)
        distances.append((row, dist))
    distances.sort(key=lambda tup: tup[1])
    neighbors = list()
    for i in range(n_neighbors):
        neighbors.append(distances[i][0])
    return np.array(neighbors)
def make_prediction(x,row_num,n_neighbors,weights):
    """
    Predicts the output value based on the k nearest neighbors relative to an input point.
    Inputs:
        x: array
            input the array of input points.
        row_num: int
            input the index(row number) of the desired point to make the prediction.
        n_neighbors: int
            input the number of neighbors to be calculated.
        weights: string
            input the type of weight to be used in the calculation. weights can be: 'uniform'(uniform weights,i.e, all points on the neighborhood are weighted equally) or 'distance'(weight points by the inverse of their distance).
    Returns:
        predictions: float
            outputs the prediction calculated based on the input point(row_num).

    """
    neighbors = get_neighbors(x,row_num,n_neighbors)
    outputs = [points[neighbors.shape[1] - 1] for points in neighbors]
    if weights == 'uniform':
        predictions = np.round(sum(outputs)/n_neighbors)
        return predictions
    elif weights == 'distance':
        for i in range(len(outputs)):
            outputs[i] = outputs[i]/(i+1)
        predictions = np.round(sum(outputs)/n_neighbors)
        return predictions
    else:
        raise ValueError("Please insert a valid weight")
class NearestNeighbors:
    """
    Class of the K-Nearest Neighbors algorithm. 
    Fits both regression and classification problems.
    This algorithm performs a brute force search, meaning that it performs poorly on large datasets, since it scales according to O[DN^2], where N is the number of samples and D is the number of dimensions.
    Methods:
        predict(x,y) -> Predict value for X.
        kneighbors(row_num,n_neighbors) -> Gets the k-nearest neighbors from x.
    """
    def __init__(self,n_neighbors = 5,weights = 'uniform'):
        """
        Initial parameters.
        Inputs:
            n_neighbors: int
                input the number of neighbors to be calculated; default is set to 5.
            weights: string
                input the type of weight to be used in the calculation. weights can be: 'uniform'(uniform weights,i.e, all points on the neighborhood are weighted equally) or 'distance'(weight points by the inverse of their distance); default is set to 'uniform'.
        """
        self.n_neighbors = n_neighbors
        self.weights = weights
    def predict(self,x,y):
        """
        Makes a prediction based on a given dataset.
        Inputs:
            x: array
                input the array of input points.
            y: array
                input the array of output points.
        Returns:
            predictions: array
                outputs the array of predictions.
        """
        self.input = x 
        self.output = y.reshape((-1,1))
        self.data = np.hstack((self.input,self.output))
        predictions = np.zeros((len(self.output)))
        for i in range(len(predictions)):
            predictions[i] = make_prediction(self.data,i,self.n_neighbors,self.weights)
        return predictions 
    def kneighbors(self,row_num,n_neighbors):
        """
        Gets the k-nearest neighbors to a certain point.
        Inputs:
            row_num: int
                input the index(row number) of the desired point to calculate the k nearest neighbors.
            n_neighbors: int
                input the number of neighbors to be calculated.     
        """ 
        kneighbors = get_neighbors(self.input,row_num,n_neighbors)
        return kneighbors


# # basicMLpy.loss_functions module
import numpy as np 

def check_inputs_shape(arr1,arr2):
	"""
	Checks if the dimensions of two arrays match and prints a warning if the dont.
	Inputs:
	    arr1: array
	        input array whose dimension is to be checked
	    arr2: array
	        input array whose dimension is to be checked      

	"""
	if (len(arr1.shape) != (len(arr2.shape))):
		print("WARNING! The shape of the prediction array is different to that of the target array, this may lead to errors.")
		print("Try passing arrays with equal shapes!")

def mse(prediction,target):
	"""
	Calculates the Mean Squared Error between two arrays
	Inputs:
	    prediction: array of shape (N,*)
	        input array of predictions
	    target: array of shape (N,*), same as prediction
	        input array of targets
	Outputs:
	    error: float or array of floats
	        outputs the mean squared error between the two arrays

	"""
	check_inputs_shape(prediction,target)
	error = np.sum(np.square(target - prediction))
	return (error/len(target))

def mae(prediction,target):
	"""
	Calculates the Mean Absolute Error between two arrays
	Inputs:
	    prediction: array of shape (N,*)
	        input array of predictions
	    target: array of shape (N,*), same as prediction
	        input array of targets
	Outputs:
	    error: float or array of floats
	        outputs the mean squared absolute between the two arrays
	"""
	check_inputs_shape(prediction,target)
	error = np.sum(np.abs(target - prediction))
	return (error/len(target))

def huber_loss(prediction,target,delta=0.5):
	"""
	Calculates the Huber Loss(also referred as Smooth L1 Loss) between two arrays
	Inputs:
	    prediction: array
	        input array of predictions
	    target: array
	        input array of targets
	    delta: float
	        input delta constant for the huber loss. default = 0.5
	Outputs:
	    huber_loss: float
	        outputs the huber loss between the two arrays
	"""
	check_inputs_shape(prediction,target)
	huber_loss = 0 
	for i in range(len(target)):
		if np.absolute(target[i] - prediction[i] ) <= delta:
			huber_loss += 0.5 * np.square(target[i] - prediction[i])
		else:
			huber_loss += (delta * np.absolute(target[i] - prediction[i])) -(0.5 * (delta**2))
	return huber_loss

def exponential_loss(prediction,target):
	"""
	Calculates the Exponential Loss between two arrays
	Inputs:
	    prediction: array
	        input array of predictions
	    target: array
	        input array of targets
	Outputs:
	    exp_loss: float
	        outputs the exponential loss between the two arrays
	"""
	check_inputs_shape(prediction,target)
	return np.mean(np.exp(- prediction * target))

def standard_accuracy(prediction,target,round_prediction=False):
	"""
	Calculates the accuracy of the predictions, given a target
	Inputs:
	    prediction: array
	        input array of predictions
	    target: array
	        input array of targets
	Output
	"""
	check_inputs_shape(prediction,target)
	counter = 0
	if(round_prediction):
		rounded_prediction = np.round(prediction)
		for i in range(len(prediction)):
			if np.absolute(rounded_prediction[i] - target[i]) == 0:
				counter = counter 
			else:
				counter += 1
		accuracy = ((np.size(target,0) - counter)/np.size(target,0))
		return accuracy
	else:
		for i in range(len(prediction)):
			if np.absolute(prediction[i] - target[i]) == 0:
				counter = counter 
			else:
				counter += 1
		accuracy = ((np.size(target,0) - counter)/np.size(target,0))
		return accuracy
		
def misclassification(y,prediction):
    """
    Calculates the number of misclassifications between the prediction and the output.
    Inputs:
        y: int
            Input a single value of the output variable y.
        prediction: int
            Input a single valut of the predicted output.
    Returns:
        misclassifications: array
            Output 1 if the values do not match and 0 if they do.
    """
    y=y.reshape((-1,1))
    prediction = prediction.reshape((-1,1))
    misclassifications = 1*(y != prediction)
    return misclassifications


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="basicMLpy", 
    version="1.0.2",
    author="Henrique Silva",
    author_email="henriquesoares@dcc.ufmg.br",
    description="A collection of simple machine learning algorithms",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/HenrySilvaCS/basicMLpy",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    py_modules=["classification","regression","model_selection","ensemble","nearest_neighbors","utils","loss_functions"],
    install_requires=[
       'numpy>=1.19',
       'scipy>=1.5.2',
       'scikit-learn>=0.23'
    ],
    python_requires='>=3.8',
)